/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource;

import org.apache.ws.addressing.EndpointReference;
import org.apache.ws.metadata.MetadataConfiguration;

/**
 * Defines a basic interface through which resources are discovered and removed. The purpose of
 * <code>ResourceHome</code> is to interact with a collection of resources of the same type. Each resource type will
 * have its own custom implementation the <code>ResourceHome</code> interface. The implementation is expected to provide
 * custom methods for creating new resources and optionally methods that act on a set of resource objects.
 */
public interface ResourceHome
   extends ResourceCreation,
           ResourceDestruction
{
   /**
    * Returns the MetadataConfigImpl containing any configured Metadata for the Service.
    *
    * @return MetadataConfigImpl
    */
   MetadataConfiguration getMetadataConfig(  );

   /**
    * Returns the name of the JSR-109 port component corresponding to this home.
    * Note, the port component name ends up being the last path component of the
    * service endpoint URL.
    *
    * @return the name of the JSR-109 port component corresponding to this home
    */
   String getPortComponentName(  );

   /**
    * Sets the classname of the Resource Impl.
    *
    * @param className The class name of the Resource Impl.
    */
   void setResourceClassName( String className );

   /**
    * Returns the Class of the Resource Impl.
    *
    * @return The Class of the Resource Impl
    */
   String getResourceClassName(  );

   /**
    * Sets the name of the SOAP header element from which this home will extract
    * a resource identifier. Set to null if this home manages a singleton resource.
    *
    * @param name resource identifier header element name, represented as
    *             a string as defined by {@link javax.xml.namespace.QName#toString()}
    *             (e.g. "{http://http://ws.apache.org/namespaces/wsrf/}ResourceIdentifier")
    */
   void setResourceIdentifierReferenceParameterName( String name );

   /**
    * Gets the name of the SOAP header element from which this home will extract
    * a resource identifier.
    *
    * @return the resource identifier header element name, represented as
    *         a string in the format defined by
    *         {@link javax.xml.namespace.QName#toString()};
    *         or null if this home manages a singleton resource
    */
   String getResourceIdentifierReferenceParameterName(  );

   /**
    * Sets the Service Impl's .Class
    *
    * @param className String representation of the class
    */
   void setServiceClassName( String className );

   /**
    * Returns the name of the Service class.
    *
    * @return the name of the Service class
    */
   String getServiceClassName(  );

   /**
    * Sets the target namespace of the WSDL of this home's service.
    *
    * @param targetNamespace
    */
   void setWsdlTargetNamespace( String targetNamespace );

   /**
    * Returns the target namespace of the WSDL.
    *
    * @return The String of the targetnamespace
    */
   String getWsdlTargetNamespace(  );

   /**
    * Adds the specified resource to this home if it is not already present
    * (optional operation).  More formally, adds the specified resource,
    * <code>resource</code>, to this home if this home contains no resource
    * <code>existing</code> such that
    * <code>(resource.getID().equals(existing.getID())</code>.
    * If this home already contains a resource with the same ID as the
    * specified resource, the call leaves the home unchanged and returns
    * <tt>false</tt>.
    *
    * This method will also check to see if the EndpointReference (EPR) has been set on
    * the Resource.  If the EPR is null, it will create one and set it on the Resource.
    *
    * @param resource the resource to be added; must not be null
    *
    * @throws UnsupportedOperationException if the <tt>add</tt> method is not
    *           supported by this home
    * @throws IllegalResourceTypeException if the resource is not of the type
    *         supported by this home
    */
   boolean add( Resource resource )
   throws IllegalResourceTypeException;

   /**
    * Extracts the Resource identifier from the specified resource context -
    * typically, this is done by deserializing a particular SOAP header
    * element that contains the identifier.
    *
    * @param context a resource context
    *
    * @return the Resource identifier
    */
   Object extractResourceIdentifier( ResourceContext context );

   /**
    * Extracts the Resource identifier from the specified endpoint reference -
    * typically, this is done by deserializing a particular reference parameter
    * that represents the identifier.
    *
    * @param epr the endpoint reference of a WS-Resource
    *
    * @return the Resource identifier
    */
   Object extractResourceIdentifier( EndpointReference epr );

   /**
    * Retrieves a resource. <b>Note:</b> This function must not return null. It must
    * either return the resource object or throw an exception if there is no resource
    * with the specified identifier.
    * </br>
    * If a resource is found and its EndpointReference member variable is not set, the
    * EndpointReference variable will get set on the resource with values obtained from
    * the jndi-config configuration.
    *
    * @param resourceId a resource identifier
    *
    * @return non-null resource
    *
    * @throws ResourceUnknownException    if no resource exists with the given key
    * @throws ResourceException           if any other error occurs.
    */
   Resource find( Object resourceId )
   throws ResourceUnknownException, 
          ResourceException;

   /**
    * Removes a resource. Invokes the resource's {@link Resource#destroy()} method
    * prior to removal.
    *
    * @throws ResourceUnknownException    if no resource exists with the given key
    * @throws ResourceException           if any other error occurs
    */
   void remove( Object resourceId )
   throws ResourceUnknownException, 
          ResourceException;
}