/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.webapp;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.PosixParser;
import org.mortbay.http.SocketListener;
import org.mortbay.jetty.Server;
import org.mortbay.util.MultiException;
import java.io.File;
import java.io.IOException;

/**
 * A standalone WSRF server that uses Jetty as the HTTP server and servlet container
 * and uses Axis as the SOAP engine.
 *
 * @author Ian Springer
 */
public class WsrfServer
{
   /** DOCUMENT_ME */
   protected static final org.apache.commons.cli.Options CMD_LINE_OPTIONS = new org.apache.commons.cli.Options(  );

   static
   {
      initCmdLineOptions(  );
   }

   /**
    * Jetty HTTP server / servlet container
    */
   private Server         m_server;
   private SocketListener m_listener   = new SocketListener(  );
   private File           m_webappPath;

   /**
    * Creates a WsrfServer for running the specified web application.
    *
    * @param webappPath the path to an Apache-WSRF-based webapp (directory or warfile)
    *
    * @throws IOException if the webapp path is invalid
    */
   public WsrfServer( File webappPath )
   throws IOException
   {
      m_webappPath    = webappPath;
      m_server        = new Server(  );
      m_server.addListener( m_listener );
      String contextName = m_webappPath.getName(  );
      m_server.addWebApplication( "/" + contextName + "/*",
                                  m_webappPath.getAbsolutePath(  ) );
   }

   /**
    * Sets the HTTP listen port.
    *
    * @param port the port number
    */
   public void setPort( int port )
   {
      m_listener.setPort( port );
   }

   /**
    * Gets the HTTP listen port.
    */
   public int getPort(  )
   {
      return m_listener.getPort(  );
   }

   /**
    * Gets the webapp path.
    */
   public File getWebappPath(  )
   {
      return m_webappPath;
   }

   /**
    * A main method for starting a server instance.
    * Usage: org.apache.ws.resource.webapp.WsrfServer [-p port] webappPath
    *
    * @param args command-line arguments
    *
    * @throws Exception on any error
    */
   public static void main( String[] args )
   throws Exception
   {
      CommandLine cmdLine = new PosixParser(  ).parse( CMD_LINE_OPTIONS, args, true );
      if ( cmdLine.getArgs(  ).length != 1 )
      {
         throw new IllegalArgumentException( "Usage: " + WsrfServer.class.getName(  ) + " [-p port] webappPath" );
      }

      File       webappPath = new File( cmdLine.getArgs(  )[0] );
      WsrfServer server = new WsrfServer( webappPath );
      if ( cmdLine.hasOption( Opts.PORT ) )
      {
         server.setPort( Integer.valueOf( cmdLine.getOptionValue( Opts.PORT ) ).intValue(  ) );
      }

      server.start(  );
   }

   /**
    * Starts the server.
    *
    * @throws MultiException
    */
   public void start(  )
   throws MultiException
   {
      m_server.start(  );
   }

   /**
    * Stops the server.
    *
    * @throws InterruptedException
    */
   public void stop(  )
   throws InterruptedException
   {
      m_server.stop(  );
   }

   private static void initCmdLineOptions(  )
   {
      CMD_LINE_OPTIONS.addOption( Opts.PORT, LongOpts.PORT, true, "server listen port" );
   }

   /**
    * DOCUMENT_ME
    */
   static interface LongOpts
   {
      /**
       * DOCUMENT_ME
       */
      String PORT = "port";
   }

   /**
    * DOCUMENT_ME
    */
   static interface Opts
   {
      /**
       * DOCUMENT_ME
       */
      String PORT = "p";
   }
}