/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util.spring;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.FileSet;
import org.apache.ws.util.JaxpUtils;
import org.apache.ws.util.XalanOutputKeys;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * An Ant task that merges one or more source spring-beans.xml's into a target spring-beans.xml.
 */
public class SpringBeansXmlMergeTask
   extends Task
{
   private List m_sourceFiles = new ArrayList(  );
   private File m_targetFile;

   /**
    * Creates a new {@link SpringBeansXmlMergeTask} object.
    */
   public SpringBeansXmlMergeTask(  )
   {
      initContextClassLoader(  );
   }

   /**
    * Adds the specified file to the list of source spring-beans.xml's.
    *
    * @param srcFile
    */
   public void setSourceFile( File srcFile )
   {
      m_sourceFiles.add( srcFile );
   }

   /**
    * Sets the target spring-beans.xml (i.e. the spring-beans.xml to be merged into).
    *
    * @param targetFile
    */
   public void setTargetFile( File targetFile )
   {
      m_targetFile = targetFile;
   }

   /**
    * Adds the files in the specified fileset to the list of source spring-beans.xml's.
    *
    * @param fileSet DOCUMENT_ME
    */
   public void addConfiguredSourceFiles( FileSet fileSet )
   {
      File             baseDir       = fileSet.getDir( getProject(  ) );
      DirectoryScanner dirScanner    = fileSet.getDirectoryScanner( getProject(  ) );
      String[]         includedFiles = dirScanner.getIncludedFiles(  );
      for ( int i = 0; i < includedFiles.length; i++ )
      {
         m_sourceFiles.add( new File( baseDir, includedFiles[i] ) );
      }
   }

   /**
    * DOCUMENT_ME
    *
    * @throws org.apache.tools.ant.BuildException
    *          DOCUMENT_ME
    */
   public void execute(  )
   throws BuildException
   {
      if ( m_sourceFiles.isEmpty(  ) )
      {
         throw new BuildException( "No source spring-beans.xml files were specified!" );
      }

      try
      {
         System.out.println( "Loading target spring-beans document from " + m_targetFile + " ..." );
         Document targetDoc       = parseSpringBeansXmlFile( m_targetFile );
         Element  targetBeansElem = targetDoc.getDocumentElement(  );
         for ( int i = 0; i < m_sourceFiles.size(  ); i++ )
         {
            mergeSpringBeans( (File) m_sourceFiles.get( i ), targetBeansElem );
         }

         saveTargetDocToFile( targetDoc );
      }
      catch ( Exception e )
      {
         throw new BuildException( e );
      }
   }

   private String getAttribute( Element elem,
                                String  attribName )
   {
      String value = elem.getAttribute( attribName );
      return ( !value.equals( "" ) ) ? value : null;
   }

   private Element getBeanElementById( Element beansElem,
                                       String  beanId )
   {
      if ( beanId == null )
      {
         return null;
      }

      Element[] beanElems = getChildElementsByTagName( beansElem, "bean" );
      for ( int i = 0; i < beanElems.length; i++ )
      {
         Element beanElem = beanElems[i];
         if ( beanId.equals( getBeanId( beanElem ) ) )
         {
            return beanElem;
         }
      }

      return null;
   }

   private String getBeanId( Element beanElem )
   {
      String id = getAttribute( beanElem, "id" );
      if ( id == null )
      {
         id = getAttribute( beanElem, "name" );
      }

      return id;
   }

   private Element[] getChildElementsByTagName( Element parentElem,
                                                String  name )
   {
      List     matchingElems = new ArrayList(  );
      NodeList childNodes = parentElem.getChildNodes(  );
      for ( int i = 0; i < childNodes.getLength(  ); i++ )
      {
         Node node = childNodes.item( i );
         if ( ( node.getNodeType(  ) == Node.ELEMENT_NODE ) && node.getNodeName(  ).equals( name ) )
         {
            matchingElems.add( node );
         }
      }

      return (Element[]) matchingElems.toArray( new Element[matchingElems.size(  )] );
   }

   private String getTextValue( Element elem )
   {
      String   textValue  = null;
      NodeList childNodes = elem.getChildNodes(  );
      if ( childNodes.getLength(  ) == 1 )
      {
         Node node = childNodes.item( 0 );
         if ( node.getNodeType(  ) == Node.TEXT_NODE )
         {
            textValue = node.getNodeValue(  );
         }
      }

      return textValue;
   }

   private Element getTopicNamespaceSetElem( Element targetBeansElem )
   {
      Element   setElem             = null;
      Element[] constructorArgElems = getChildElementsByTagName( targetBeansElem, "constructor-arg" );
      if ( constructorArgElems.length == 1 )
      {
         Element[] setElems = getChildElementsByTagName( constructorArgElems[0], "set" );
         if ( setElems.length == 1 )
         {
            setElem = setElems[0];
         }
      }

      return setElem;
   }

   private void mergeSpringBeans( File    srcFile,
                                  Element targetBeansElem )
   throws Exception
   {
      System.out.println( "Merging " + srcFile + " into target spring-beans document ..." );
      Document  srcDoc       = parseSpringBeansXmlFile( srcFile );
      Element   srcBeansElem = srcDoc.getDocumentElement(  );
      Element[] beanElems    = getChildElementsByTagName( srcBeansElem, "bean" );
      for ( int i = 0; i < beanElems.length; i++ )
      {
         Element srcBeanElem         = beanElems[i];
         String  srcBeanId           = getBeanId( srcBeanElem );
         Element existingBeanElem    = getBeanElementById( targetBeansElem, srcBeanId );
         Node    importedSrcBeanElem = targetBeansElem.getOwnerDocument(  ).importNode( srcBeanElem, true );
         if ( existingBeanElem != null )
         {
            if ( srcBeanId.equals( "TopicNamespaceRegistry" ) )
            {
               updateTopicNamespaceRegistryBeanDef( srcBeanElem, existingBeanElem );
            }
            else
            {
               System.out.println( "Replacing existing bean element with id '" + srcBeanId + "' ..." );
               targetBeansElem.replaceChild( importedSrcBeanElem, existingBeanElem );
            }
         }
         else
         {
            System.out.println( "Adding new bean element with id '" + srcBeanId + "' ..." );
            targetBeansElem.appendChild( importedSrcBeanElem );
         }
      }
   }

   private Document parseSpringBeansXmlFile( File springBeansXmlFile )
   throws Exception
   {
      return JaxpUtils.loadDocument( new FileInputStream( springBeansXmlFile ) );
   }

   private void saveTargetDocToFile( Document doc )
   throws Exception
   {
      System.out.println( "Saving updated target spring-beans document to " + m_targetFile + " ..." );
      FileOutputStream targetOutStream = new FileOutputStream( m_targetFile );
      targetOutStream.write( toString( doc ).getBytes(  ) );
      targetOutStream.close(  );
   }

   private static String toString( Node node )
   throws Exception
   {
      TransformerFactory transformerFactory = TransformerFactory.newInstance(  );
      Transformer        transformer = transformerFactory.newTransformer(  );
      transformer.setOutputProperty( OutputKeys.DOCTYPE_PUBLIC, "-//SPRING//DTD BEAN//EN" );
      transformer.setOutputProperty( OutputKeys.DOCTYPE_SYSTEM,
                                     "http://www.springframework.org/dtd/spring-beans.dtd" );
      transformer.setOutputProperty( OutputKeys.INDENT, "yes" );
      transformer.setOutputProperty( XalanOutputKeys.OUTPUT_PROP_INDENT_AMOUNT, "2" );
      DOMSource    domSource    = new DOMSource( node );
      StringWriter stringWriter = new StringWriter(  );
      StreamResult result       = new StreamResult( stringWriter );
      transformer.transform( domSource, result );
      return ( stringWriter.toString(  ) );
   }

   private void initContextClassLoader(  )
   {
      if ( Thread.currentThread(  ).getContextClassLoader(  ) == null )
      {
         Thread.currentThread(  ).setContextClassLoader( Task.class.getClassLoader(  ) );
      }
   }

   private void updateTopicNamespaceRegistryBeanDef( Element srcBeanElem,
                                                     Element existingBeanElem )
   {
      System.out.println( "Merging topic namespace values into existing TopicNamespaceRegistry bean definition..." );
      Element   srcSetElem         = getTopicNamespaceSetElem( srcBeanElem );
      Element[] srcValueElems      = getChildElementsByTagName( srcSetElem, "value" );
      Element   targetSetElem      = getTopicNamespaceSetElem( existingBeanElem );
      Element[] existingValueElems = getChildElementsByTagName( targetSetElem, "value" );
      Set       existingValues     = new HashSet(  );
      for ( int j = 0; j < existingValueElems.length; j++ )
      {
         existingValues.add( getTextValue( existingValueElems[j] ) );
      }

      for ( int j = 0; j < srcValueElems.length; j++ )
      {
         Element srcValueElem = srcValueElems[j];
         if ( !existingValues.contains( getTextValue( srcValueElem ) ) )
         {
            Node importedSrcValueElem = targetSetElem.getOwnerDocument(  ).importNode( srcValueElem, true );
            targetSetElem.appendChild( importedSrcValueElem );
         }
      }
   }
}